
static const u32 RingBuffer_Size=8*1024; // req power of 2
static const u32 RingBuffer_DstSampleRate=44100;
typedef struct {
  u32 SrcSampleRate,SamplesPerFlame;
  volatile u32 ReadIndex,WriteIndex;
  u32 *pbuf;
  u32 freqcur,freqadd;
  s16 freqL0,freqR0,freqL1,freqR1;
  u32 *pFlameBuf;
} TRingBuffer;

static TRingBuffer RingBuffer;

static void audioRingBufferCallback(void* buf, unsigned int length, void *userdata);
static bool audioRingBufferUpdate(u32 (*pFunc_Update)(u32*));

static volatile bool audioRingBufferInterrupt;

static void audioRingBufferInit(void)
{
  TRingBuffer *prb=&RingBuffer;
  
  prb->SrcSampleRate=0;
  prb->SamplesPerFlame=0;
  
  prb->ReadIndex=0;
  prb->WriteIndex=0;
  
  prb->pbuf=(u32*)malloc(RingBuffer_Size*4);
  for(u32 idx=0;idx<RingBuffer_Size;idx++){
    prb->pbuf[idx]=0;
  }
  
  prb->pFlameBuf=NULL;
  
  prb->freqcur=0;
  prb->freqadd=0;
  
  prb->freqL0=0;
  prb->freqR0=0;
  prb->freqL1=0;
  prb->freqR1=0;
  
  audioRingBufferInterrupt=false;
  
  pspAudioSetVolume(0, PSP_VOLUME_MAX,PSP_VOLUME_MAX);
  pspAudioSetChannelCallback(0, audioRingBufferCallback, NULL);
}

static void audioRingBufferSetVolume15(u32 Volume15)
{
  pspAudioSetVolume(0, Volume15,Volume15);
}

static void audioRingBufferStart(u32 SrcSampleRate,u32 SamplesPerFlame,u32 (*pFunc_Update)(u32*))
{
  TRingBuffer *prb=&RingBuffer;
  
  if(prb->SrcSampleRate!=SrcSampleRate){
    while(1){
      u32 ridx=prb->ReadIndex;
      const u32 widx=prb->WriteIndex;
      u32 last=(RingBuffer_Size+widx-ridx)&(RingBuffer_Size-1);
      if(last==0) break;
      sceDisplayWaitVblankStart();
    }
    while(1){
      if(audioRingBufferInterrupt==false) break;
      sceDisplayWaitVblankStart();
    }
  }
  
  if((RingBuffer_Size-2048)<SamplesPerFlame){
    printf("Flame size overflow. (%d-2048)<%d\n",RingBuffer_Size,SamplesPerFlame);
    SystemHalt();
  }
  
  prb->SrcSampleRate=SrcSampleRate;
  prb->SamplesPerFlame=SamplesPerFlame;
  
  prb->pFlameBuf=(u32*)malloc(prb->SamplesPerFlame*4);
  
  prb->freqcur=0;
  prb->freqadd=prb->SrcSampleRate*0x10000/RingBuffer_DstSampleRate;
  
  prb->freqL0=0;
  prb->freqR0=0;
  prb->freqL1=0;
  prb->freqR1=0;
  
  while(1){
    if(audioRingBufferUpdate(pFunc_Update)==false) break;
  }
}

static void audioRingBufferEnd(void)
{
  TRingBuffer *prb=&RingBuffer;
  
  if(prb->pFlameBuf!=NULL){
    free(prb->pFlameBuf); prb->pFlameBuf=NULL;
  }
}

static void audioRingBufferFree(void)
{
  TRingBuffer *prb=&RingBuffer;
  
  pspAudioSetChannelCallback(0, NULL, NULL);
  
  while(1){
    if(audioRingBufferInterrupt==false) break;
    sceDisplayWaitVblankStart();
  }
  
  if(prb->pbuf!=NULL){
    free(prb->pbuf); prb->pbuf=NULL;
  }
}

static void audioRingBufferCallback(void* buf, unsigned int length, void *userdata)
{
  TRingBuffer *prb=&RingBuffer;
  
  audioRingBufferInterrupt=true;
  
  const u32 buflength=length;
  u32 *pbuf32=(u32*)buf;
  
  if(prb->pbuf==NULL){
    length=0;
    }else{
    u32 ridx=prb->ReadIndex;
    const u32 widx=prb->WriteIndex;
    
    u32 last=(RingBuffer_Size+widx-ridx)&(RingBuffer_Size-1);
    if(last<length) length=last;
    
    u32 cur=prb->freqcur;
    const u32 add=prb->freqadd;
    s16 L0=prb->freqL0;
    s16 R0=prb->freqR0;
    s16 L1=prb->freqL1;
    s16 R1=prb->freqR1;
    
    if(add==0x10000){
      for(u32 idx=0;idx<length;idx++){
        u32 smp=prb->pbuf[ridx];
        ridx=(ridx+1)&(RingBuffer_Size-1);
        pbuf32[idx]=smp;
      }
      }else{
      for(u32 idx=0;idx<length;idx++){
        cur+=add;
        while(0x10000<cur){
          cur-=0x10000;
          u32 smp=prb->pbuf[ridx];
          prb->pbuf[ridx]=0;
          ridx=(ridx+1)&(RingBuffer_Size-1);
          L0=L1; R0=R1;
          L1=smp&0xffff; R1=smp>>16;
        }
        s16 L=((L0*(0x10000-cur))+(L1*cur))/0x10000;
        s16 R=((R0*(0x10000-cur))+(R1*cur))/0x10000;
        pbuf32[idx]=(L&0xffff)|(R<<16);
      }
    }
    
    prb->freqcur=cur;
    prb->freqL0=L0;
    prb->freqR0=R0;
    prb->freqL1=L1;
    prb->freqR1=R1;
    
    prb->ReadIndex=ridx;
  }
  
  for(u32 idx=length;idx<buflength;idx++){
    pbuf32[idx]=0;
  }
  
  audioRingBufferInterrupt=false;
}

static bool audioRingBufferUpdate(u32 (*pFunc_Update)(u32*))
{
  TRingBuffer *prb=&RingBuffer;
  
  { // Check ring buffer.
    const u32 ridx=prb->ReadIndex;
    const u32 widx=prb->WriteIndex;
    u32 req=(RingBuffer_Size+ridx-widx)&(RingBuffer_Size-1);
    if(req==0) req=RingBuffer_Size;
    if(req<=prb->SamplesPerFlame) return(false);
  }
  
  u32 *pBufLR=prb->pFlameBuf;
  
  u32 SamplesCount=pFunc_Update(pBufLR);
  if(prb->SamplesPerFlame<SamplesCount){
    printf("Flame buffer overflow. %d<%d.\n",prb->SamplesPerFlame,SamplesCount);
  }
  if(SamplesCount==0) return(false);
  
  const u32 ridx=prb->ReadIndex;
  u32 widx=prb->WriteIndex;
  
  for(u32 idx=0;idx<SamplesCount;idx++){
    prb->pbuf[widx]=pBufLR[idx];
    widx=(widx+1)&(RingBuffer_Size-1);
  }
  
  prb->WriteIndex=widx;
  
  return(true);
}

